use std::ops::{Add, AddAssign, Div, DivAssign, Mul, MulAssign, Neg, Sub, SubAssign};

use crate::{Float, Numeric, Primitive};

/// Struct representing a complex number.
///
/// # Example
///
/// ```rust
/// # use lineal::{Complex, Float, Numeric};
/// let c = Complex { real: 9.0, imag: 0.0 };
/// assert_eq!(format!("{:?}", c), "Complex { real: 9.0, imag: 0.0 }");
///
/// let three = c.fsqrt().real;
/// assert_eq!(three, 3.0);
///
/// let i: Complex<f32> = (-Complex::whole(1)).fsqrt();
/// assert_eq!(i, Complex::i());
///
/// let a = Complex { real: 2.0, imag: 0.5 };
/// let a_squared = a * a;
/// assert_eq!(a_squared, Complex { real: 3.75, imag: 2.0 });
/// assert_eq!(a_squared.fsqrt(), a);
/// ```
#[derive(Debug, Copy, Clone)]
pub struct Complex<T: Float + Numeric + Primitive> {
    pub real: T,
    pub imag: T,
}

impl<T: Float + Numeric + Primitive> Complex<T> {
    /// The imaginary unit.
    pub fn i() -> Self {
        Self {
            real: T::whole(0),
            imag: T::whole(1),
        }
    }
}

impl<T: Float + Numeric + Primitive> Add for Complex<T> {
    type Output = Self;

    fn add(self, rhs: Self) -> Self::Output {
        Self {
            real: self.real + rhs.real,
            imag: self.imag + rhs.imag,
        }
    }
}

impl<T: Float + Numeric + Primitive> AddAssign for Complex<T> {
    fn add_assign(&mut self, rhs: Self) {
        self.real += rhs.real;
        self.imag += rhs.imag;
    }
}

impl<T: Float + Numeric + Primitive> Sub for Complex<T> {
    type Output = Self;

    fn sub(self, rhs: Self) -> Self::Output {
        Self {
            real: self.real - rhs.real,
            imag: self.imag - rhs.imag,
        }
    }
}

impl<T: Float + Numeric + Primitive> SubAssign for Complex<T> {
    fn sub_assign(&mut self, rhs: Self) {
        self.real -= rhs.real;
        self.imag -= rhs.imag;
    }
}

impl<T: Float + Numeric + Primitive> Mul for Complex<T> {
    type Output = Self;

    fn mul(self, rhs: Self) -> Self::Output {
        Self {
            real: self.real * rhs.real - self.imag * rhs.imag,
            imag: self.real * rhs.imag + self.imag * rhs.real,
        }
    }
}

impl<T: Float + Numeric + Primitive> MulAssign for Complex<T> {
    fn mul_assign(&mut self, rhs: Self) {
        *self = *self * rhs;
    }
}

impl<T: Float + Numeric + Primitive> Div for Complex<T> {
    type Output = Self;

    fn div(self, rhs: Self) -> Self::Output {
        let divisor = rhs.real * rhs.real + rhs.imag * rhs.imag;
        let mut result = self * Self { real: rhs.real, imag: -rhs.imag };
        result.real /= divisor;
        result.imag /= divisor;
        return result;
    }
}

impl<T: Float + Numeric + Primitive> DivAssign for Complex<T> {
    fn div_assign(&mut self, rhs: Self) {
        *self = *self / rhs;
    }
}

impl<T: Float + Numeric + Primitive> Neg for Complex<T> {
    type Output = Self;

    fn neg(self) -> Self::Output {
        Self {
            real: -self.real,
            imag: -self.imag,
        }
    }
}

impl<T: Float + Numeric + Primitive> PartialEq for Complex<T> {
    fn eq(&self, other: &Self) -> bool {
        self.real == other.real && self.imag == other.imag
    }

    fn ne(&self, other: &Self) -> bool {
        self.real != other.real || self.imag != other.imag
    }
}

impl<T: Float + Primitive + Numeric> Float for Complex<T> {
    fn fabs(self) -> Self {
        Self {
            real: (self.real * self.real + self.imag * self.imag).fsqrt(),
            imag: T::whole(0),
        }
    }

    fn fsqrt(self) -> Self {
        let zero = T::whole(0);
        let two = T::whole(2);
        let abs_z = self.fabs().real;

        let base_re = if self.imag == zero { self.real } else { (abs_z + self.real) / two };
        let re = if base_re < zero { zero } else { base_re.fsqrt() };

        let im = if self.imag == zero { zero } else {
            (self.imag / self.imag.fabs()) * ((abs_z - self.real) / two).fsqrt()
        } + if base_re < zero { (-base_re).fsqrt() } else { zero };

        Self {
            real: re,
            imag: im,
        }
    }

    fn fsin(self) -> Self {
        Self {
            real: self.real.fsin() * self.imag.fcosh(),
            imag: self.real.fcos() * self.imag.fsinh(),
        }
    }

    fn fcos(self) -> Self {
        Self {
            real: self.real.fcos() * self.imag.fcosh(),
            imag: self.real.fsin() * self.imag.fsinh(),
        }
    }

    fn fsinh(self) -> Self {
        Self {
            real: self.real.fsinh() * self.imag.fcos(),
            imag: self.real.fcosh() * self.imag.fsin(),
        }
    }

    fn fcosh(self) -> Self {
        Self {
            real: self.real.fcosh() * self.imag.fcos(),
            imag: self.real.fsinh() * self.imag.fsin(),
        }
    }
}

impl<T: Float + Numeric + Primitive> Numeric for Complex<T> {
    fn whole(value: u32) -> Self {
        Self { real: T::whole(value), imag: T::whole(0) }
    }
}

#[cfg(test)]
mod tests {
    use crate::{Complex, Float};

    #[test]
    fn division() {
        let mut a: Complex<f32> = Complex { real: 3.0, imag: 2.0 };
        let b: Complex<f32> = Complex { real: 4.0, imag: -5.0 };
        let expected: Complex<f32> = Complex { real: 2.0 / 41.0, imag: 23.0 / 41.0 };
        a /= b;
        assert_eq!(format!("{:?}", a), format!("{:?}", expected));
    }

    #[test]
    fn square_root_with_negatives() {
        let a = Complex { real: -1.0, imag: -1.0 };
        let root = a.fsqrt();
        let expected = Complex { real: 0.45508986, imag: -1.09868411 };
        assert!((root.real - expected.real).fabs() < 0.00001);
        assert!((root.imag - expected.imag).fabs() < 0.00001);
    }
}
